
#  LabTools - uncertainties.py
#  Copyright 2019 Luca Arnaboldi

import numpy
import math
import uncertainties as unc
from uncertainties import unumpy
from scipy.optimize import curve_fit

from functools import wraps

def unarray(data, u_data):
    """
    Crea a numpy.array of ufloat given two arrays of float.

    This function is now deprecated since it's already implemented in unumpy.
    This is now only a redirect to unumpy mantained for retrocompability.
    """

    if len(data) != len(u_data):
        raise IndexError('Two arrays have different leght: {0} and {1}'.format(
        len(data),
        len(u_data),
    ))

    # Old Version
    # return numpy.array([unc.ufloat(data[i], u_data[i]) for i in range(0, len(data))])

    return unumpy.uarray(data, u_data)


def unpack_unarray(u_array):
    """
    Divide an array of ufloats in two numpy.array of floats.
    It works also for unumpy.uarray and tuple.
    """
    data = numpy.zeros(len(u_array), dtype = 'float64')
    u_data = numpy.zeros(len(u_array), dtype = 'float64')

    for i in range(0, len(u_array)):
        data[i] = u_array[i].n
        u_data[i] = u_array[i].s

    return data, u_data


def de2unc(value, dig, percent = 0., quad = True):
    # Double Error to uncertainties
    """
    Given a value with its digit error and percentage error return the uncertaties
    element.
    If quad is true the error are summated in quadrature.
    """
    err_perc = abs(value * percent / 100)
    if quad:
        error = numpy.sqrt(dig**2 + err_perc**2)
    else:
        error = abs(dig) + err_perc

    # Try with floats
    try:
        return unc.ufloat(value, error)
    # Use arrays
    except AttributeError:
        return unarray(value, error)

def ucurve_fit(f, xdata, ydata, **kwargs):
    """
    Wrapper for curve_fit that allows use of uncertainties both in model and data.
    It returns a tuple o ufloat for parametres, correlated.
    """

    x, ux = unpack_unarray(xdata)
    y, uy = unpack_unarray(ydata)

    uy = unc.covariance_matrix(ydata) # I use the covariance matrix as sigma, to fit correlated values

    @wraps(f) # need for pass the number of parametres. Without curve_fit fails
    def model(x, *pars):
        # if you give to nominal_values a standard  numpy.array it returns it
        return unumpy.nominal_values(f(x, *pars))

    p, cov = curve_fit(
        f = model,
        xdata = x,
        ydata = y,
        sigma = uy,
        **kwargs
    )

    return unc.correlated_values(p, cov)




from LabTools.latex import Variable, Decimal, UDecimal, Document

from uncertainties import ufloat

def test_variable():
    # Basic methods testing
    v = Variable("name", "value")
    assert(v.latexcode() == "\\newcommand{\\name}{value}")
    v.setname("n4m3")
    assert(v.latexcode() == "\\newcommand{\\n4m3}{value}")
    v.setvalue("v4lu3")
    assert(v.latexcode() == "\\newcommand{\\n4m3}{v4lu3}")
    
def test_decimal():
    d = Decimal('test', 45)
    assert(d.latexcode() == "\\newcommand{\\test}{45.00}")
    d.setdigits(7)
    assert(d.latexcode() == "\\newcommand{\\test}{45.00000}")
    d.setvalue(450000100)
    assert(d.latexcode() == "\\newcommand{\\test}{4.500001e+8}")

def test_udecimal_starndard():
    ud = UDecimal("R", 3423.34768, unc = 8.2385, unc_digit = 2)
    assert(ud.latexcode() == "\\newcommand{\\R}{3423.3 \\pm 8.2 e0}")
    ud.setuncdigit(3)
    assert(ud.latexcode() == "\\newcommand{\\R}{3423.35 \\pm 8.24 e0}")
    ud.setvalue(8.436)
    ud.setuncdigit(2)
    ud.setunc(0.056)
    assert(ud.latexcode() == "\\newcommand{\\R}{8.436 \\pm 0.056 e0}")

    ud1 = UDecimal("test", 87, unc = 7, unc_digit = 2)
    try:
        assert(ud1.latexcode() == "\\newcommand{\\test}{87.0 \\pm 7.0 e0}")
    except Warning:
        pass
    ud1.setuncdigit(1)
    assert(ud1.latexcode() == "\\newcommand{\\test}{87 \\pm 7 e0}")
    ud1.setunc(7.0)
    assert(ud1.latexcode() == "\\newcommand{\\test}{87 \\pm 7 e0}")
    ud1.setuncdigit(4)
    try:
        assert(ud1.latexcode() == "\\newcommand{\\test}{87.000 \\pm 7.000 e0}")
    except Warning:
        pass
        
def test_udecimal_uncertainties():
    ud = UDecimal("test", ufloat(3423.34768, 8.2385))
    assert(ud.latexcode() == "\\newcommand{\\test}{3423 \\pm 8 e0}")
    
    a = ufloat(3.456, 0.056)
    b = ufloat(4.98, 0.000003)

    ud.setvalue(a + b)
    assert(ud.latexcode() == "\\newcommand{\\test}{8.44 \\pm 0.06 e0}")
    ud.setvalue(a*b)
    assert(ud.latexcode() == "\\newcommand{\\test}{17.2 \\pm 0.3 e0}")
    
    c = ufloat(0.0e-3, 1.3e-3)
    ud.setvalue(c)
    ud.setuncdigit(2)
    assert(ud.latexcode() == "\\newcommand{\\test}{0.0000 \\pm 0.0013 e0}")
    
    
    

def check_file_with_document(d, filename):
    with open(filename) as f:
        cnt = 0
        for l in f:
            if l.startswith('%') or l.startswith('\n'):
                continue
            assert(d.variables[cnt].latexcode() + '\n' == l)
            cnt = cnt + 1

def test_document():
    # Basic method
    d = Document()
    d.setvariable(Variable('name1', 'value1'))
    assert(d.variables[0].name == 'name1' and d.variables[0].value == 'value1')
    d.setvariable(Variable('name2', 'value2'))
    d.setvariable(Variable('name2', 'value3'))
    assert(d.variables[1].name == 'name2' and d.variables[1].value == 'value3')
    
    # Fail inserting not Variable
    try:
        d.setvariable("string")
    except TypeError:
        pass
        
    # Saving
    d.save('test.tmp')
    check_file_with_document(d, 'test.tmp')
    
    d.clearvariables()
    assert(len(d.variables) == 0)
    d.setvariable(Decimal("R", 3423.34768, 6))
    d.save('test.tmp')
    check_file_with_document(d, 'test.tmp')
    
    
    d.setvariable(UDecimal("R", ufloat(4.57,0.3)))
    d.setvariable(UDecimal("a", ufloat(4.54,0.32), unc_digit = 2))
    d.setvariable(UDecimal("b", ufloat(4.55,0.31), unc_digit = 2))
    d.setvariable(UDecimal("S", ufloat(4.56,0.30)))
    
    # Deleting
    d.removevariable("R")
    assert(len(d.variables) == 3)
    try:
        d.removevariable("R")
    except Warning:
        pass
    d.removevariable("S")
    assert(len(d.variables) == 2)
    assert(d.variables[0].latexcode() == "\\newcommand{\\a}{4.54 \\pm 0.32 e0}")
    assert(d.variables[1].latexcode() == "\\newcommand{\\b}{4.55 \\pm 0.31 e0}")




#  LabTools - latex.py
#  Copyright 2019 Luca Arnaboldi


from .utils import significant_digits, move_decimal, most_significant_digit
from .utils import pair_decimal_with_uncertainty, unpack_unarray
from .utils import percentual_error_digit
import numpy

DEFAULT_SIGNIFICANT_DIGITS = 4
DEFAULT_UNC_DIGITS = 2
DEFAULT_UNC_PERCENT_ACCURACY = 25
FILE_HEADER = '''
% File generated with LabTools.latex
% Copyright (c) Luca Arnaboldi 2019

'''


################################################################################
# Tables

class TabularColumn():
    
    def __init__(
        self,
        array,
        show_unc=None,
        siunitx_num=None,
        global_unc=None,
        unc_digits=None,
        value_digits=None,
        unc_perc_accuracy=None
    ):
        try:
            self.array, self.unc = unpack_unarray(array)
        except AttributeError:
            self.array = array
            self.unc = numpy.array([None] * len(array))
        self.show_unc = show_unc
        self.siunitx_num = siunitx_num
        self.global_unc = global_unc
        self.unc_digits = unc_digits
        self.value_digits = value_digits
        self.unc_perc_accuracy = unc_perc_accuracy
    

class TabularContent():
    
    def __init__(
        self,
        hlines=False, 
        show_unc=True,
        siunitx_num=True,
        unc_digits=None,
        value_digits=DEFAULT_SIGNIFICANT_DIGITS,
        unc_perc_accuracy = DEFAULT_UNC_PERCENT_ACCURACY,
    ):
        self.hlines = hlines
        self.show_unc = show_unc
        self.siunitx_num = siunitx_num
        self.unc_digits = unc_digits
        self.value_digits = value_digits
        self.unc_perc_accuracy = unc_perc_accuracy
        self.columns = []
        
    def add_column(self, *args, **kwargs):
        if args[0].__class__.__name__ == 'ndarray':
            self.columns.append(TabularColumn(args[0], **kwargs))
        elif args[0].__class__.__name__ == 'TabularColumn':
            self.columns.append(args[0])
        else:
            raise(TypeError('TabularContent.add_column() does not support this arguments.'))
        
        
    def save(self, filename):
        tex = FILE_HEADER
        tex += '% Include this code in tabular enviroment to display this content\n'
        
        # Get the number of rows
        content_lines = []
        number_of_rows = 0
        for c in self.columns:
            number_of_rows = max(number_of_rows, len(c.array))
        
        # build the content lines
        content_lines = [""] * number_of_rows
        
        # Insert elements
        for i in range(0, len(self.columns)):
            c = self.columns[i]
            cnt = 0 # Count how many row in this column
            
            ## Setting up column
            # show_unc
            show_unc = self.columns[i].show_unc
            if show_unc == None:
                show_unc = self.show_unc
            # siunitx_num
            siunitx_num = self.columns[i].siunitx_num
            if siunitx_num == None:
                siunitx_num = self.siunitx_num
            # unc_digits
            unc_digits = self.columns[i].unc_digits
            if unc_digits == None:
                unc_digits = self.unc_digits
            # value_digits
            value_digits = self.columns[i].value_digits
            if value_digits == None:
                value_digits = self.value_digits
            # unc_perc_accuracy
            unc_perc_accuracy = self.columns[i].unc_perc_accuracy
            if unc_perc_accuracy == None:
                unc_perc_accuracy = self.unc_perc_accuracy
            
            
            for j in range(0, len(c.array)):
                cnt += 1 # Add 1 to the row counter
                
                # There's no unc for this column (neither gloabal or local)
                if c.global_unc == None and c.unc.all() == None:
                    if siunitx_num == False:
                        raise(NotImplementedError('siunitx_num = False'))
                    else:
                        content_lines[j] += '\\num{{{0}}}'.format(
                            significant_digits(c.array[j], value_digits, True)
                        )
                # There's unc
                else:
                    if siunitx_num == False:
                        raise(NotImplementedError('siunitx_num = False'))
                    else:
                        # If there's global_unc
                        if c.unc.all() == None:
                            act_unc = c.global_unc
                        else:
                            act_unc = c.unc[j]
                            
                        # Compute the number of digit of the uncertainty of this
                        # element.
                        if unc_digits == None:
                            unc_digits = percentual_error_digit(
                                act_unc,
                                unc_perc_accuracy,
                            )
                        assert(not unc_digits == None)
                        value_s, unc_s, e, warn = pair_decimal_with_uncertainty(
                            c.array[j],
                            act_unc,
                            unc_digits,
                        )
                        content_lines[j] += '\\num{{{0}'.format(value_s)
                        if show_unc:
                            content_lines[j] += ' \\pm {0}'.format(unc_s)
                        content_lines[j] += ' e{0}}}'.format(e)
                        
                # Next element. Last column & is removed next
                content_lines[j] += ' &'

            # Fill the empty lines of the table
            while cnt < number_of_rows:
                content_lines[cnt] += ' &'
                cnt += 1
                    
                    
                

        # End lines and attach to the tex
        for l in content_lines:
            l = l[:-1] # Remove & char at the end of the line
            l += ' \\\\'
            if self.hlines:
                l += ' \\hline'
            tex += l + '\n'
        
        # Write file
        with open(filename, 'w') as tfile:
            tfile.write(tex)



################################################################################
# Variables

class Variable:
    def __init__(self, name, value):
        self.name = name
        self.value = value

    def setvalue(self, value):
        self.value = value

    def setname(self, name):
        self.name = name

    def latexcode(self):
        return '\\newcommand{{\\{0}}}{{{1}}}'.format(
            self.name,
            str(self.value),
        )


class Decimal(Variable):
    """
        It needs Latex Package siunitx to work properly.
    """
    def __init__(self, name, value, digits=DEFAULT_SIGNIFICANT_DIGITS):
        super().__init__(name, float(value))
        self.digits = digits

    def setdigits(self, digits):
        self.digits = digits

    def latexcode(self, ):
        return '\\newcommand{{\\{0}}}{{{1}}}'.format(
            self.name,
            significant_digits(self.value, self.digits, True),
        )


class UDecimal(Variable):
    """
    It needs Latex Package siunitx to work properly.
    """

    def __init__(self, name, value, unc_digit=None, unc_perc_accuracy = DEFAULT_UNC_PERCENT_ACCURACY, unc = None):
        # uncertainties.ufloat passed
        if unc is None:
            value_ = value.nominal_value
            unc_ = value.std_dev
        # other case
        else:
            value_ = value
            unc_ = unc
            
        super().__init__(name, value_)
        self.unc = float(unc_)
        if unc_digit is None:
            self.unc_digit = percentual_error_digit(
                self.unc,
                unc_perc_accuracy,
            )
        else:
         self.unc_digit = int(unc_digit)

    def setunc(self, unc):
        self.unc = float(unc)

    def setuncdigit(self, unc_digit):
        self.unc_digit = int(unc_digit)
        
    def setvalue(self, value):
        try:
            self.value = value.nominal_value
            self.unc = value.std_dev
        except AttributeError:
            self.value = float(value)

    def latexcode(self):
        to_print_value, to_print_unc, unc_digit_rep, warn = pair_decimal_with_uncertainty(
            self.value,
            self.unc,
            self.unc_digit,
        )
        
        # Raise a warning if adding digits to uncertanty
        if warn:
            raise Warning("""uncertanty of Udecimal '{0}' has more digit than
                             orginal value: {1} -- {2}""".format(self.name,
                                                                 self.unc,
                                                                 to_print_unc))

        return '\\newcommand{{\\{0}}}{{{1} \\pm {2} e{3}}}'.format(
            self.name,
            to_print_value,
            to_print_unc,
            unc_digit_rep,
        )


class Document:
    def __init__(self):
        self.variables = []

    def setvariable(self, variable):
        # Check correct type
        if issubclass(variable.__class__, Variable):
            # Try to replace an existence one
            for i in range(0, len(self.variables)):
                if self.variables[i].name == variable.name:
                    self.variables[i] = variable
                    return
            # Adding if not present
            self.variables.append(variable)
        else:
            raise TypeError('object {0} ({1}) is not a valid variable.'.format(
                variable,
                type(variable),
            ))

    def removevariable(self, name):
        for i in range(0, len(self.variables)):
            if self.variables[i].name == name:
                del self.variables[i]
                return
        raise Warning('failed removing {0}. Variable {0} not in Document'.format(
            name,
        ))

    def clearvariables(self):
        self.variables.clear()

    def save(self, filename):
        # Header message
        tex = FILE_HEADER
        tex += '% Include this file in your LaTex project to use these definitions\n\n'

        # Variables 
        tex += '% Variables\n'
        for v in self.variables:
            tex += v.latexcode() + '\n'
        
        with open(filename, 'w') as tfile:
            tfile.write(tex)




from LabTools.utils.uncertainties import *
from LabTools.plot import *

## Generic stuff for simulate a fit
def model(x, a, b):
    return numpy.sqrt(a * x + b)
        
def umodel(x, a, b):
    return unumpy.sqrt(a * x + b)

x = numpy.array([3., 41., 100.])
y = numpy.array([5., 15., 25.])
ux = x / 100.
uy = y / 100.
    
X = unarray(x, ux)
Y = unarray(y, uy)

fitted = ucurve_fit(model, X, Y)




def test_residual_plot():
    # Generic
    residual_plot(umodel, fitted, X, Y, use_ux = False)
    
    #PDF
    residual_plot(umodel, fitted, X, Y,
                  use_ux = False, figfile = 'test.tmp.pdf',
                  title = '\\LaTeX')
    #TKZ: actually it does not work :( 
    try:
        residual_plot(umodel, fitted, X, Y, use_ux = False, figfile = 'test.tmp.tex')
    except NotImplementedError:
        pass
        
def test_errorbars_plot():
    errorbars_plot(X,Y, xlogscale = True, ylogscale = True, figfile = 'errorbarstest.tmp.pdf')




from LabTools.fit import *

from LabTools.utils import unarray
import numpy
from uncertainties import unumpy, ufloat

# Import the testing of ucurve_fit from uncertainties tests.
from test_uncertainties import test_ucurve_fit as test_ucurve_fit_

"""
This is not testing the correcteness of the fit or the error propagation of
uncertainties. It is only checking if it works in every case I need.
"""
def model(x, a, b):
    return numpy.sqrt(a * x + b)

def umodel(x, a, b):
    return unumpy.sqrt(a * x + b)

def dumodel(x, a, b):
    return a / (2 * unumpy.sqrt(a * x + b))

x = numpy.array([3., 41., 100.])
y = numpy.array([5., 15., 25.])
ux = x / 100.
uy = y / 100.

X = unarray(x, ux)
Y = unarray(y, uy)

p0 = ucurve_fit(umodel, X, Y)

def test_ucurve_fit_with_chi2():
    print(chi2(umodel, p0, X, Y))

def test_iteraed_fit_with_chi2():
    p = iterated_fit(umodel, X, Y, df = dumodel)
    assert((p[0]-p0[0]).n < 0.01)
    assert((p[1] - p0[1]).n < 0.01)
    print(chi2iterated(umodel, p, X, Y, df = dumodel))

    # Not Coverge
    try:
        p = iterated_fit(umodel, X, Y, df = dumodel, max_iterations = 0)
    except Warning:
        pass

    # Numeric derivative
    try:
        p = iterated_fit(umodel, X, Y)
    except NotImplementedError:
        pass
    try:
        chi2iterated(umodel, p, X, Y)
    except NotImplementedError:
        pass


def test_ucurve_fit():
    test_ucurve_fit_()

def test_circle_fit_paper():
    """
    This test was extracted from the paper: "Least-Squares Circle" Fit by Randy Bullock.

    It test correctness of the fit.
    """
    x = numpy.array([0.000, 0.500, 1.000, 1.500, 2.000, 2.500, 3.000])
    y = numpy.array([0.000, 0.250, 1.000, 2.250, 4.000, 6.250, 9.000])
    xc, yc, r = circle_fit(x,y)
    assert(xc + 11.839 < 0.001)
    assert(yc - 8.4464 < 0.0001)
    assert(r  - 14.686 < 0.001)

    # This is to check if it works also with uncertainties values
    X = unarray(x, [0.001]*len(x))
    Y = unarray(y, [0.001]*len(y))
    Xc, Yc, R = circle_fit(X,Y)

    # test the chi squared quantity
    assert(relative_difference_circle_fit(x,y,xc,yc,r)==relative_difference_circle_fit(X,Y,Xc,Yc,R))




#  LabTools - fit.py
#  Copyright 2019 Luca Arnaboldi

from .utils import ucurve_fit, unpack_unarray
import uncertainties as unc
from uncertainties import unumpy
from uncertainties import umath
from scipy.misc import derivative
import numpy
from scipy.optimize import curve_fit

from functools import wraps

DEFAULT_CONVERGENCE_CONDITION = 1e-7
DEFAULT_MAX_ITERATION = 20


def chi2(f, param, X, Y):
    """
    Calculate the chi squared and the degree of freedom of a standard fit.
    X, Y must be uarrays!
    """
    @wraps(f)
    def model(x, *pars):
        return unumpy.nominal_values(f(x, *pars))
    x, ux = unpack_unarray(X)
    y, uy = unpack_unarray(Y)
    return sum(((y - model(x, *param)) / uy)**2), len(X) - len(param)

def iterated_fit(
    f,
    X,
    Y,
    p0 = None,
    df = None,
    absolute_sigma = False,
    convergence_condition = DEFAULT_CONVERGENCE_CONDITION,
    max_iterations = DEFAULT_MAX_ITERATION,
    ):
    """
    Iterated fit that propagates the errors on the X on the Y to account the x-axis
    error.
    If p0 is None is calculated with ucurve_fit without X error.
    If df is None it is used the numeric derivative from scipy. Not always work.
    """

    if df is None:
        raise NotImplementedError("Numeric derivatived not implemented yet!")
    else:
        @wraps(df)
        def dmodel(x, *pars):
            return unumpy.nominal_values(df(x, *pars))

    @wraps(f)
    def model(x, *pars):
        return unumpy.nominal_values(f(x, *pars))

    x, ux = unpack_unarray(X)
    y, uy = unpack_unarray(Y)

    # Now we can work with normal numpy functions, without unceraties

    par, cov = curve_fit(model, x, y, sigma = uy, p0 = p0, absolute_sigma = absolute_sigma)

    while --max_iterations:
        dif_y = numpy.sqrt(uy**2 + (dmodel(x, *par) * ux)**2)

        npar, ncov = curve_fit(
            model,
            x,
            y,
            sigma = dif_y,
            p0 = par,
            absolute_sigma = absolute_sigma
        )

        perror = numpy.abs(npar - par) / npar
        cerror = numpy.abs(ncov - cov) / ncov

        if (perror < convergence_condition).all() and (cerror < convergence_condition).all():
            break

        par = npar
        cov = ncov

    if max_iterations == 0:
        raise Warning("iterated_fit termined because max_iterations was reached.")

    return unc.correlated_values(par, cov)

def chi2iterated(f, param, X, Y, df = None):
    """
    Calculate the chi squared and degree of freedom taking to account the error
    on x axis.
    """
    if df is None:
        raise NotImplementedError("Numeric derivatived not implemented yet!")
    else:
        @wraps(df)
        def dmodel(x, *pars):
            return unumpy.nominal_values(df(x, *pars))
    @wraps(f)
    def model(x, *pars):
        return unumpy.nominal_values(f(x, *pars))

    x, ux = unpack_unarray(X)
    y, uy = unpack_unarray(Y)
    dif_y = numpy.sqrt(uy**2 + (ux * dmodel(x, *param))**2) # errore
    return sum(((y - model(x, *param)) / dif_y)**2), len(X) - len(param)

def circle_fit(X, Y):
    """
    Implementation of a Least-Squares Circle Fit as described by Randy Bullock in circle_fit.pdf.

    It returns coordinates of the center and radius.
    """
    assert(len(X) == len(Y))
    N = len(X)

    xm = sum(X)/ N
    ym = sum(Y)/N

    u = X - xm
    v = Y - ym

    Suuu = sum(u**3)
    Svvv = sum(v**3)
    Suvv = sum(u*(v**2))
    Suuv = sum((u**2)*v)
    Suu = sum(u**2)
    Svv = sum(v**2)
    Suv = sum(u*v)

    # Can't find an easy way to solve the system with uncertaintis. So I did it by hand:
    # a1 x + b1 y = c1
    # a2 x + b2 y = c2
    a1 = Suu
    b1 = Suv
    c1 = (Suuu + Suvv)/2
    a2 = Suv
    b2 = Svv
    c2 = (Svvv + Suuv)/2

    det = a1*b2-b1*a2
    uc = -(b1*c2-c1*b2)/det
    vc = (a1*c2-c1*a2)/det

    # Solutions
    xc = uc + xm
    yc = vc + ym
    radius = umath.sqrt(uc**2 + vc**2 + (Suu + Svv)/N)

    return xc, yc, radius

def relative_difference_circle_fit(x, y, xc, yc, radius):
    """
    Calculate the totL distance of points from the circle, normalized with the number of points and the radius.
    """
    assert(len(x) == len(y))
    try:
        x_, trash = unpack_unarray(x)
        y_, trash = unpack_unarray(y)
        r_ = radius.n
        xc_ = xc.n
        yc_ = yc.n
    except AttributeError:
        x_ = x
        y_ = y
        r_ = radius
        xc_ = xc
        yc_ = yc

    return sum(numpy.sqrt((x_-xc_)**2+(y_-yc_)**2))/r_/len(x_)




from LabTools.measure import *
import numpy as np


def test_instrument():
    i = Instrument('test/inputs/instrument1.yaml')
    assert(str(i.measure('tipo1', 15.)) == '15+/-10')
    assert(str(i.measure('tipo1', 74.6)) == '75+/-12')
    assert(str(i.measure('tipo1', 0.34)) == '0.34+/-0.05')
    
    assert(str(i.measure('tipo2', 1.876)) == '1.876+/-0.031')
    assert(str(i.measure('tipo2', 1.876, 5.000)) == '1.876+/-0.021')
    
    # Tipo che non esiste
    try:
        i.measure('tipononesiste', 0.)
    except KeyError:
        pass
        
    # Provo a passare un value che non è un float
    try:
        i.measure('tipo1', 'uno')
    except ValueError:
        pass
        
    # Fuori il fondoscala più grande
    try:
        i.measure('tipo1', 5000.)
    except ValueError:
        pass
        
    # Fondoscala che non esiste
    try:
        i.measure('tipo1', 5., fond = 50.)
    except ValueError:
        pass
        
def test_tester():
    t = Tester('test/inputs/multimetro_digitale.yaml')
    v = np.array([2.87, 89.7])
    t.voltage(v)
    
def test_oscilloscope():
    o = Oscilloscope('test/inputs/oscilloscopio.yaml')
    
    v = np.array([2.87, 89.7e-2])
    o.voltage(v)
    
    t = np.array([1.6e-6, 12.e-3])
    o.time(t)
    o.frequency(1./t)
    
    




from LabTools.utils.generic import *
from uncertainties import ufloat, unumpy
import cv2
import filecmp

def float_equal(x, y, accuracy):
    return (1. - accuracy) <= x / y and x / y  <= (1. + accuracy)

def image_equal(image1, image2):
    return image1.shape == image2.shape and not(np.bitwise_xor(image1,image2).any())


def test_sprint(capfd):
    a = 'marco'
    b = 45.
    c = [12, 45]
    d = ufloat(b,b)
    e = unumpy.uarray(c, c)

    sprint(a)
    out, err = capfd.readouterr()
    assert(out == 'a: marco\n')

    sprint(b)
    out, err = capfd.readouterr()
    assert(out == 'b: 45.0\n')

    sprint(c)
    out, err = capfd.readouterr()
    assert(out == 'c: [12, 45]\n')

    sprint(d)
    out, err = capfd.readouterr()
    assert(out == 'd: (4+/-4)e+01\n')

    sprint(e)
    out, err = capfd.readouterr()
    assert(out == 'e: [12.0+/-12.0 45.0+/-45.0]\n')

def test_decibel():
    assert(float_equal(decibel(0.5), -6.0205999132796239042747778944898, 1e-6))
    assert(float_equal(decibel(10.), 20., 1e-6))

    # With uncertainties
    a = ufloat(10., 1.)
    assert(str(decibel(a)) == "20.0+/-0.9")

def test_crop_oscilloscope_image():
    crop_oscilloscope_image('test/inputs/to_crop.png', 'test.tmp.png')

    correct = cv2.imread('test/outputs/cropped.png')
    produced = cv2.imread('test.tmp.png')
    print(correct - produced)

    assert(image_equal(correct, produced))

def test_conversion_deg_rad():
    assert(deg2rad(1.) == 0.017453292519943295)
    assert(deg2rad(0., 60.) == 0.017453292519943295)
    assert(deg2rad(0., 0., 3600.) == 0.017453292519943295)

    assert(rad2deg(0.017453292519943295) == 1.)




#  LabTools - plot.py
#  Copyright 2019 Luca Arnaboldi

from .utils import unpack_unarray
import matplotlib.pyplot as plot
import numpy
from uncertainties import unumpy

from functools import wraps

##  Costant
DEFAULT_PLOT_DIMENSION = [10., 7.5]
DEFAULT_ERRORBAR_STYLE = {
    'linestyle' : '',
    'c' : 'k',
    'capsize' : 0,
    'elinewidth' : 0.7,
}
DEFAULT_PLOT_STYLE = {
    'c' : 'b',
    'lw' : .3,
}
DEFAULT_NORM_RES_STYLE = {
    'marker' : '.',
    'c' : 'b',
    'markersize' : 3.,
    'linestyle' : '',
}

DEFAULT_OUTLIERS_STYLE = {
    'linestyle' : '',
    'c' : 'red',
    'capsize' : 0,
    'elinewidth' : 0.7,
}

DEFAULT_OUTLIERS_RES_STYLE = {
    'marker': 'x',
    'linestyle' : '',
    'c' : 'red',
    'markersize' : 3.,
}

DEFAULT_SECOND_PLOT_STYLE = {
    'c' : 'g',
    'lw' : .3,
}

def rcConfig(
    usetex = True,
    latex_adds = True,
    fontsize = 12,
    ):
    """
    This defines the rcConfig for my plots.
    Thanks to: Niccolò Porciani
    """
    # Set the usage of LaTeX in labels and titles. It needs a LaTeX compiler
    # installed on the PC.
    plot.rcParams['text.usetex'] = usetex
    # LaTeX things needed:
    #   - siunitx
    #   - amsmath
    if latex_adds:
        plot.rcParams['text.latex.preamble'] = [
            r'''
            \usepackage{amsmath}
            \usepackage{siunitx}
            '''
        ]
    plot.rcParams['font.family'] = 'serif'
    plot.rcParams['font.serif'] = 'Computer Modern'
    plot.rcParams['font.size'] = fontsize

def xlimit(x, ux, logscale):
    if logscale:
        d_interval = (numpy.log10(max(x)) - numpy.log10(min(x))) * 0.01
        return (
            numpy.power(10., numpy.log10(min(x)) - d_interval),
            numpy.power(10., numpy.log10(max(x)) + d_interval)
        )
    else:
        d_interval = (max(x) - min(x)) * 0.01
        return (
            min(x) - max(ux) - d_interval,
            max(x) + max(ux) + d_interval
        )


def savepdf(fig, figfile):
    if figfile is not None:
        if figfile.endswith('tex'):
            raise NotImplementedError('tikzplotlib does not suppurt subplots.')
            """
            save_tikz(
                figure = fig,
                figurewidth = '\\linewidth',
                filepath = figfile,
                textsize = fontsize,
            )
            """
        else:
            fig.savefig(figfile, format = 'pdf', bbox_inches = 'tight')

def multi_plot(
    data,
    xlabel = "",
    ylabel = "",
    title = None,
    fontsize = 12,
    xlogscale = False,
    ylogscale = False,
    figfile = None,
    ):
    """
    Plot many set of points
    """
    # Load Configuration
    rcConfig(fontsize = fontsize)

    fig = plot.figure()
    fig.set_size_inches(*DEFAULT_PLOT_DIMENSION)
    main = plot.subplot2grid((12,9), (0,0), colspan = 9, rowspan = 12, fig = fig)

    for point_set in data:
        main.plot(point_set[0], point_set[1])

    main.minorticks_on()

    if xlogscale:
        main.set_xscale('log')

    if ylogscale:
        main.set_yscale('log')

    main.set_xlabel(xlabel)
    main.set_ylabel(ylabel)

    if title is not None:
        main.set_title(title)

    ## Salva immagine
    savepdf(fig, figfile)


def errorbars_plot(
    X,
    Y,
    xlabel = "",
    ylabel = "",
    title = None,
    fontsize = 12,
    xlogscale = False,
    ylogscale = False,
    figfile = None,
    ):
    """
    Plot points with errorbars
    """
    # Load Configuration
    rcConfig(fontsize = fontsize)

    x, ux = unpack_unarray(X)
    y, uy = unpack_unarray(Y)

    fig = plot.figure()
    fig.set_size_inches(*DEFAULT_PLOT_DIMENSION)
    main = plot.subplot2grid((12,9), (0,0), colspan = 9, rowspan = 12, fig = fig)

    main.set_xlim(xlimit(x, ux, xlogscale))

    main.minorticks_on()
    main.errorbar( x, y, xerr = ux, yerr = uy, **DEFAULT_ERRORBAR_STYLE)

    if xlogscale:
        main.set_xscale('log')

    if ylogscale:
        main.set_yscale('log')

    main.set_xlabel(xlabel)
    main.set_ylabel(ylabel)

    if title is not None:
        main.set_title(title)

    ## Salva immagine
    savepdf(fig, figfile)



def residual_plot(
    f,
    param,
    X,
    Y,
    use_ux = True,
    df = None,
    xlabel = "",
    normres = True,
    fontsize = 12,
    ylabel = "",
    title = None,
    rylabel = None,
    xlogscale = False,
    ylogscale = False,
    outliers = None,
    figfile = None,
    second_f = None,
    #second_df = None,
    second_param = None,
    ):
    """
    It takes X and Y as unumpy.uarray and makes a plot with the residual graph.
    """

    # Load Configuration
    rcConfig(fontsize = fontsize)

    # Redefine f  and df for curvefit
    @wraps(f) # need for pass the number of parametres. Without, curve_fit fails
    def uf(x, *pars):
        # if you give to nominal_values a standard  numpy.array it returns it
        return unumpy.nominal_values(f(x, *pars))
    # df is analogous
    if df is not None:
        @wraps(df)
        def udf(x, *pars):
            return unumpy.nominal_values(df(x, *pars))

    # Processing the second function
    if second_f is not None:
        @wraps(second_f)
        def suf(x, *pars):
            return unumpy.nominal_values(second_f(x, *pars))
    """
    if second_df is not None:
        @wraps(second_df)
        def sudf(x, *pars):
            return unumpy.nominal_values(second_f(x, *second_pars))
    """



    x, ux = unpack_unarray(X)
    y, uy = unpack_unarray(Y)
    pars, upars = unpack_unarray(param)

    fig = plot.figure()
    fig.set_size_inches(*DEFAULT_PLOT_DIMENSION)
    main = plot.subplot2grid((12,9), (0,0), colspan = 9, rowspan = 6, fig = fig)
    res = plot.subplot2grid((12,9), (6,0), colspan = 9, rowspan = 3, fig = fig)

    fig.subplots_adjust(hspace=0) # Non lascia spazio tra i grafici

    ## Calcola residui
    #Propago errori su y da x
    if use_ux:
        if df is None:
            raise TypeError('Errors on x were given but no derivative of f')
        else:
            errore = numpy.sqrt( uy**2 + (udf(x, *pars) * ux)**2 )
    else:
        errore = uy
    residui = y - uf(x, *pars)
    if normres:
        residui = residui / errore

    # Adding the outliers if not present
    if outliers is None:
        outliers = [False] * len(X)


    ## Setto i limiti sulle x
    xlim = xlimit(x, ux, xlogscale)
    main.set_xlim(xlim)
    res.set_xlim(xlim)

    ##Crea la grid appropriata e setta i limiti
    if xlogscale:
        d_interval = (numpy.log10(max(x)) - numpy.log10(min(x))) * 0.01
        grid = numpy.logspace(
            numpy.log10(min(x)) - d_interval,
            numpy.log10(max(x)) + d_interval,
            1000
        )
    else:
        d_interval = (max(x) - min(x)) * 0.01
        grid = numpy.linspace(
            min(x) - max(ux) - d_interval,
            max(x) + max(ux) + d_interval
        )

    # Plot dei punti
    for i in range(0, len(X)):
        err_bar_style = DEFAULT_ERRORBAR_STYLE
        norm_res_style = DEFAULT_NORM_RES_STYLE
        if outliers[i]:
            err_bar_style = DEFAULT_OUTLIERS_STYLE
            norm_res_style = DEFAULT_OUTLIERS_RES_STYLE

        main.errorbar(x[i], y[i], xerr = ux[i], yerr = uy[i], **err_bar_style)
        if normres:
            res.plot(x[i], residui[i], **norm_res_style)
        else:
            res.errorbar(x[i], residui[i], xerr = 0., yerr = uy[i], **err_bar_style)

    # Grafico curva
    #main.grid()
    main.minorticks_on()
    main.plot(grid, uf(grid, *pars), **DEFAULT_PLOT_STYLE)
    if second_f is not None:
        main.plot(grid, suf(grid, *second_param), **DEFAULT_SECOND_PLOT_STYLE)

    if xlogscale:
        main.set_xscale('log')

    if ylogscale:
        main.set_yscale('log')
    main.set_xticklabels([]) #Leva i numeri sotto il primo grafico

    main.set_ylabel(ylabel)
    if title is not None:
        main.set_title(title)

    ## Grafico residui (solo impostazioni, i punti li ho già messi sopra)
    #res.grid()
    res.minorticks_on()
    res.plot(grid, grid * 0., **DEFAULT_PLOT_STYLE)


    if xlogscale:
        res.set_xscale('log')

    res.set_xlabel(xlabel)

    if rylabel is None:
        if normres:
            res.set_ylabel('Res. Norm')
    else:
        res.set_ylabel(rylabel)



    ## Salva immagine
    savepdf(fig, figfile)




__pkgname__ = 'labtools-arna'
__version__ = '1.0.0'
__author__ = 'Luca Arnaboldi'
__license__ = 'GPLv3'
__summary__ = 'Tools for physics laborayory courses'
__url__ = 'https://github.com/arn4/LabTools'




#  LabTools - generic.py
#  Copyright 2019 Luca Arnaboldi

import inspect
from PIL import Image
from uncertainties import unumpy as unp
import numpy as np

def sprint(obj):
    """
    Super Print. Print a variable with its name.
    """
    def retrieve_name(var):
        """
        Copy-pasted from StackOverflow
        """
        callers_local_vars = inspect.currentframe().f_back.f_back.f_locals.items()
        return [var_name for var_name, var_val in callers_local_vars if var_val is var]

    obj_name = retrieve_name(obj)[0]

    print("{0}: {1}".format(obj_name, obj))

def decibel(x):
    """
    Convert a value in decibel.
    """
    return 20. * unp.log10(x)

def deg2rad(deg, prim = None, sec = None):
    """
    Convert a value from degree to radiants
    """
    if sec is not None:
        rad_sec = sec / 3600. * np.pi / 180.
    else:
        rad_sec = 0.
    if prim is not None:
        rad_min = prim / 60. * np.pi / 180.
    else:
        rad_min = 0.
    rad_deg = deg * np.pi / 180.
    return rad_deg+rad_min+rad_sec

def rad2deg(rad):
    """
    Convert radiants to degrees (not supported the notation with minute and seconds)
    """
    return rad/np.pi * 180.


def crop_oscilloscope_image(image, result_image = None, area = None):
    """
    Crop an acquisition of oscilloscope screen and save it in result_image.
    If result_image is None, the image is overwritten.
    Border define the interesting area. If None the default is used.
    """
    AREA = (0, 23, 317, 245)

    if result_image is None:
        result_image = image
    if area is None:
        area = AREA

    img = Image.open(image)
    cropped_img = img.crop(AREA)

    cropped_img.save(result_image)




#  LabTools - utils
#  Copyright 2019 Luca Arnaboldi

from .significant_digits import significant_digits, move_decimal
from .significant_digits import most_significant_digit, pair_decimal_with_uncertainty
from .significant_digits import percentual_error_digit

from .uncertainties import unarray, unpack_unarray, de2unc, ucurve_fit

from .generic import sprint, decibel, crop_oscilloscope_image, deg2rad, rad2deg




#  LabTools - measure.py
#  Copyright 2019 Luca Arnaboldi

from .utils import de2unc

import numpy
from operator import attrgetter
import yaml

class Instrument():
    """
    This class handle a generic misuration instrument.
    The configuration must be given in a configuration file in YAML format.
    """
    
    def __init__(self, config):
        with open(config) as conf:
            self.measure_types = yaml.full_load(conf)
        
        # Sort all the measure types by ascending order of scale
        for measure_type, scale in self.measure_types.items():
            scale.sort(key = lambda scale: scale['full-scale'])
            
    def measure_single(self, measure_type, value, fond = None):
        """
        Take a value measured with this instrument and it returns an uncertainty
        item. The error is calculated with the specification for this instrument
        given in the configuration.
        If fond is None the best full-scale one is choosed.
        """

        ref = float(value)
        
        for item in self.measure_types[measure_type]:
            if (fond is None and value < item['full-scale']) or (fond is not None and float(fond) == float(item['full-scale'])):
                return de2unc(
                    value,
                    item['resolution'] * item['digit_error'],
                    item['percentage_error'],
                )
        # Right full-scale not found        
        raise ValueError('Unable to compute this measure: {0}, {1}, {2}'.format(
            measure_type,
            value,
            fond
        ))
        
    def measure_array(self, measure_type, data, fond = None):
        
        if fond is not None:
            if len(data) != len(fond):
                raise IndexError('Two arrays have different leght: {0} and {1}'.format(
                len(data),
                len(fond),
            ))
        else:
            fond = [None] * len(data)
            
        return numpy.array(
            [self.measure_single(measure_type, data[i], fond[i]) for i in range(0, len(data))]
        )
        
    def measure(self, measure_type, value, fond = None):
        try:
            return self.measure_single(measure_type, value, fond)
        except TypeError:
            return self.measure_array(measure_type, value, fond)
            
        

class Tester(Instrument):
    """
    Class specialized for the multimeter.
    Voltage are in V.
    Currents are in mA.
    Capacitance are in nF.
    Resistance are in Ohm.
    """
    
    def voltage(self, value, fond = None, AC = False): 
        if AC:
            measure_type = 'ACtension'
        else:
            measure_type = 'DCtension'
        return self.measure(measure_type, value, fond)
    
    def current(self, value, fond = None, AC = False):
        if AC:
            measure_type = 'ACcurrent'
        else:
            measure_type = 'DCcurrent'
        return self.measure(measure_type, value, fond)
         
    def resistance(self, value, fond = None):
        return self.measure('resistance', value, fond)
        
    def capacitance(self, value, fond = None):
        return self.measure('capacitance', value, fond)
        
    def frequency(self, value, fond = None):
        return self.measure('frequency', value, fond)
        
    def temperature(self, value, fond = None):
        return self.measure('temperature', value, fond)


class Oscilloscope(Instrument):
    
    def voltage(self, value, fond = None):
        return self.measure('voltage', value, fond)
    
    def time(self, value, fond = None):
        return self.measure('time', value, fond)
    
    def trigger_frequency(self, value):
        return self.measure('trigger_frequency', value)
        
    def time_frequency(self, value, fond = None):
        return self.measure('frequency', value, fond)
    
    def frequency(self, value, fond = None):
        # The frequency measure is given by the time
        return 1./ self.time_frequency(1./value, fond)
        
    





#  LabTools - significant_digit
#  Copyright 2019 Luca Arnaboldi

import math


def significant_digits(x, p, compact = False):
    """
    Code heavily ispired from https://github.com/randlet/to-precision
    
    Convert x in a string written with p significant digit and the  exponent of the 
    exponential notation.
    If compact is True it retuens only a string. Example:
    0.00345, 2, False --> ("3.4", -3)
    0.00345, 2, True  --> "3.4e-3"
    """
    x = float(x)

    if x == 0.:
        ans = "0." + "0"*(p-1)
        if not compact:
            return ans, 0
        else:
            return ans

    out = []

    if x < 0:
        out.append("-")
        x = -x

    e = int(math.log10(x))
    tens = math.pow(10, e - p + 1)
    n = math.floor(x/tens)

    if n < math.pow(10, p - 1):
        e = e - 1
        tens = math.pow(10, e - p+1)
        n = math.floor(x / tens)

    if abs((n + 1.) * tens - x) <= abs(n * tens -x):
        n = n + 1

    if n >= math.pow(10,p):
        n = n / 10.
        e = e + 1

    m = "%.*g" % (p, n)
    
    exp_notation = False

    if e < -2 or e >= p:
        out.append(m[0])
        if p > 1:
            out.append(".")
            out.extend(m[1:p])

        if compact:
            out.append('e')
            if e > 0:
                out.append("+")
            out.append(str(e))
        else:
            exp_notation = True
    elif e == (p - 1):
        out.append(m)
    elif e >= 0:
        out.append(m[:e+1])
        if e+1 < len(m):
            out.append(".")
            out.extend(m[e+1:])
    else:
        out.append("0.")
        out.extend(["0"]*-(e+1))
        out.append(m)
        
    if exp_notation:
        exponent = e
    else:
        exponent = 0

    if compact:
        return "".join(out)
    else:
        return "".join(out), exponent



def most_significant_digit(x):
    """
    Return the exponent of themost significant digit of a number x
    """
    if x == 0.:
        return 0
    return int(math.floor(math.log10(abs(x))))


def move_decimal(number, delta):
    """
    Move the decimal separator of delta digits.
    """
    float(number) # Check that number is representing a float
    
    if delta == 0:
        return number
        
    p = number.find('.')
    
    if p == -1:
        number += '.' # This will not affect the orginal argument
        p = len(number) - 1
    
    if delta > 0:
        ans = number[:p]
        if p + 1 + delta < len(number):
            ans += number[p + 1:p + 1 + delta]
            ans += "."
            ans += number[p + 1 + delta:len(number)]
        else:
            ans += number[p + 1:len(number)]
            ans += "0" * (p + 1 + delta - len(number))
        return ans
    else:
        if p + delta <= 0:
            ans = "0."
            ans += "0" * (-(p + delta))
            ans += number[:p] + number[p + 1:len(number)]
        else:
            ans = number[:p + delta]
            ans += "."
            ans += number[p + delta : p]
            ans += number[p + 1 : len(number)]
        return ans 

def pair_decimal_with_uncertainty(value, unc, digits):
    """
    Given a two float value, unc, a number and its uncertanty, and digits the number
    of significant difits of the uncertanty, it returns a tuple containg (a, b, e),
    where:
    - a: is a string for value;
    - b: is a string for unc;
    - e: the common exponent.
    - warn: it is if the number of digits of the uncertainty is increased
    Strings are formatted in a way that (a +- b) 10^e is the correct representation
    of the numbers.
    """
    # Calculate the position of most significant digit of value and unc
    unc_digit = most_significant_digit(unc)
    value_digit = most_significant_digit(value)

    # Get the value and unc with the appropiate number of significant digits
    to_print_unc, unc_digit_rep = significant_digits(unc, digits)
    to_print_value, value_digit_rep = significant_digits(
        value,
        max(value_digit - unc_digit + digits, 0)
    )
    
    # return a warning if adding digits to uncertanty
    if len(str(unc)) < len(to_print_unc):
        warn = True
    else:
        warn = False

    # Move the unc value until it has the same exponent of the value
    to_print_unc = move_decimal(
        to_print_unc,
        unc_digit_rep - value_digit_rep
    )
    
        
    return to_print_value, to_print_unc, value_digit_rep, warn
    
    
def percentual_error_digit(value, percentage):
    """
    Given value and a percentage return the number of significant digits of value
    needed for specify the value up to percentage%.
    """
    
    d = 1
    approx = float(significant_digits(value, d, True))
    while abs(value - approx)/min(approx, value) > float(percentage)/100.:
        d += 1
        approx = float(significant_digits(value, d, True))
        
    return d
        




from LabTools.utils.uncertainties import *

from uncertainties import ufloat, correlated_values



def test_unarray_unpackuarray():
    a = numpy.array([ufloat(2, 3), ufloat(4, 5), ufloat(6., 7.)])
    b = numpy.array([2, 4, 6.])
    c = numpy.array([3, 5, 7.])
    d = numpy.array([3., 2.])

    b_, c_ = unpack_unarray(a)
    assert(b.all() == b_.all())
    assert(c.all() == c_.all())

    a_ = unarray(b, c)
    # We need str beacuse uncertainties behavior with comparsion
    assert(str(a_.all()) == str(a.all()))

    assert(str(a.all()) == str(unarray(*unpack_unarray(a)).all()))
    b__, c__ = unpack_unarray(unarray(b, c))
    assert(b__.all() == b.all())
    assert(c__.all() == c.all())
    try:
        unarray(a, d)
    except IndexError:
        pass

def test_de2unc():
    # Test Lab3
    assert(str(de2unc(1.673, 0.001, 0.5)) == '1.673+/-0.008')
    assert(str(de2unc(1.673, 0.001, 0.5, False)) == '1.673+/-0.009')

    assert(str(de2unc(0.167, 0.001, 0.5)) == '0.1670+/-0.0013')
    assert(str(de2unc(0.167, 0.001, 0.5, False)) == '0.1670+/-0.0018')

    # Misura negativa
    assert(str(de2unc(-1.673, 0.001, 0.5)) == '-1.673+/-0.008')
    assert(str(de2unc(-1.673, 0.001, 0.5, False)) == '-1.673+/-0.009')

    a = numpy.array([1.0, 2.0])
    b = numpy.array([0.1, 0.2])
    c = numpy.array([1, 1])
    res = numpy.array([ufloat(1.0, 0.100498), ufloat(2.00, 0.2009)])

    assert(str(de2unc(a, b, c).all()) == str(res.all()))

def test_ucurve_fit():
    """
    This is not testing the correcteness of the fit or the error propagation of
    uncertainties. It is only checking if it works in every case I need.
    """
    def model(x, a, b):
        return numpy.sqrt(a * x + b)

    def umodel(x, a, b):
        return unumpy.sqrt(a * x + b)


    x = numpy.array([3., 41., 100.])
    y = numpy.array([5., 15., 25.])
    ux = x / 100.
    uy = y / 100.

    X = unarray(x, ux)
    Y = unarray(y, uy)


    fitted1 = ucurve_fit(model, X, Y)
    fitted2 = ucurve_fit(umodel, X, Y)

    assert(str(fitted1[0]) == str(fitted2[0]))
    assert(str(fitted1[1]) == str(fitted2[1]))

def test_ucurve_fit_correlation():
    def model(x, a, b):
        return numpy.sqrt(a * x + b)
    def umodel(x, a, b):
        return unumpy.sqrt(a * x + b)

    x = numpy.array([3., 41., 100.])
    ux = x / 100.

    y = numpy.array([5., 15., 25.])
    uy = numpy.array([[0.01, 0.001,  0.01],
                      [0.001,  0.01, 0.02],
                      [0.01, 0.02, 0.05]])

    p, cov = curve_fit(model, x, y, sigma=uy)
    fitted1 = correlated_values(p, cov)
    X = unarray(x, ux)
    Y = correlated_values(y, uy)
    fitted2 = ucurve_fit(umodel, X, Y)

    assert(str(fitted1[0]) == str(fitted2[0]))
    assert(str(fitted1[1]) == str(fitted2[1]))




import setuptools
import pathlib

from LabTools  import (__pkgname__ as PKG_NAME, __version__ as VERSION,
                       __author__ as AUTHOR, __license__ as LICENSE,
                       __summary__ as SUMMARY, __url__ as URL)

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

setuptools.setup(
    name = PKG_NAME,
    version = VERSION,
    author  =  AUTHOR,
    author_email = 'luca@arnaboldi.lu',
    description = SUMMARY,
    long_description = README,
    long_description_content_type = 'text/markdown',
    url = URL,
    license = LICENSE,
    packages = setuptools.find_packages(),
    python_requires = '>=3.6',
    install_requires = [
        'numpy',
        'uncertainties',
        'matplotlib',
        'scipy',
        'pyaml'
    ]
)




from LabTools.latex import TabularContent, TabularColumn
from LabTools.utils import unarray
import filecmp, numpy

a_v = numpy.array([2., 3.])
a_u = numpy.array([0.2, 0.3])
b_v = numpy.array([2., 3., 4.])
b_u = numpy.array([0.2, 0.3, 0.4])
    

def test_empty_file_tabular():
    tc = TabularContent()
    tc.save('test.tmp')
    assert(filecmp.cmp('test/outputs/empty_tabular.tex', 'test.tmp'))
    
def test_basic_tabular():
    # Basic arrays for twsting
    tc = TabularContent()
    c1 = TabularColumn(a_v)
    c2 = TabularColumn(b_v)
    tc.add_column(c1)
    tc.add_column(c2)
    tc.save('test.tmp')
    assert(filecmp.cmp('test/outputs/basic_table.tex', 'test.tmp'))
    
def test_column():
    tc = TabularContent()
    c1 = TabularColumn(a_v, global_unc = 6.0)
    c2 = TabularColumn(unarray(b_v, b_u), show_unc = False, unc_digits = 2)
    c3 = TabularColumn(a_v, value_digits = 1)
    c4 = TabularColumn(unarray(a_v, a_u), unc_digits = 2)
    c5 = TabularColumn(unarray(b_v, b_u), unc_digits=3)
    tc.add_column(c1)
    tc.add_column(c2)
    tc.add_column(c3)
    tc.add_column(c4)
    tc.add_column(c5)
    tc.save('test.tmp')
    assert(filecmp.cmp('test/outputs/test_column.tex', 'test.tmp'))
    
    tc = TabularContent()
    tc.add_column(b_v, global_unc = 0.00044528, unc_digits = 4)
    try:
        tc.add_column('error')
    except TypeError:
        pass
        
    tc.save('test.tmp')
    assert(filecmp.cmp('test/outputs/test_column2.tex', 'test.tmp'))

def test_siunitx_num():
    tc = TabularContent()
    tc.add_column(a_v, siunitx_num = False)
    try:
       tc.save('test.tmp') 
    except NotImplementedError:
        pass
        
    tc = TabularContent()
    tc.add_column(unarray(a_v, a_u), siunitx_num = False)
    try:
       tc.save('test.tmp') 
    except NotImplementedError:
        pass
    
def test_tabular_global():
    tc = TabularContent(hlines = True)
    tc.add_column(a_v)
    tc.add_column(b_v)
    tc.save('test.tmp')
    assert(filecmp.cmp('test/outputs/tabular_hlines.tex', 'test.tmp'))
